/*
AVR Matrix
 
Author: David Voswinkel
Email:  david@optixx.org
*/

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <termios.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <gtk/gtk.h>
#include <xmms/configfile.h>
#include <xmms/plugin.h>


#define VERSION       "0.0.1"
#define NUM_BANDS     8
#define HEIGHT        8
#define HISTORY_BAR   5
#define HISTORY_PEAK  5
static GtkWidget *configure_win = NULL;
static GtkWidget *vbox, * options_frame, * options_vbox;
static GtkWidget *options_line;
static GtkWidget *options_freq;
static GtkWidget *options_invert;
static GtkWidget *options_flip;
static GtkWidget *options_dev;
static GtkWidget *options_peak;
static GtkWidget *bbox, *ok, *cancel;


static int fd = -1;
static void avr_matrix_init(void);
static void avr_matrix_cleanup(void);
static void avr_matrix_about(void);
static void avr_matrix_configure(void);
static void avr_matrix_playback_start(void);
static void avr_matrix_playback_stop(void);
static void avr_matrix_render_freq(gint16 freq_data[2][256]);
static void avr_matrix_read_config(void);

static gint bar[NUM_BANDS][HISTORY_BAR+1];
static gint peak[NUM_BANDS][HISTORY_PEAK+2];

static gdouble scale;
gint xscale1[] = {1,3,7,14,28,54,101,137,187,255};
gint xscale2[] = {1,2,3,5,7,10,14,20,28};

typedef struct
{
    gboolean mode_bar;
    gboolean mode_freq;
    gboolean mode_invert;
    gboolean mode_flip;
    gboolean mode_hold_peak;
    gchar    * device;
} avr_matrix_config_t;

avr_matrix_config_t avr_matrix_config;

VisPlugin avr_matrix_vtable = {
    0, // Handle, filled in by xmms
    0, // Filename, filled in by xmms
    0,                     // Session ID
    "AvrMatrix Plugin",       // description
    0, // # of PCM channels for render_pcm()
    1, // # of freq channels wanted for render_freq()
    avr_matrix_init,           // Called when plugin is enabled
    avr_matrix_cleanup,        // Called when plugin is disabled
    avr_matrix_about,          // Show the about box
    avr_matrix_configure,      // Show the configure box
    0,                     // Called to disable plugin, filled in by xmms
    avr_matrix_playback_start, // Called when playback starts
    avr_matrix_playback_stop,  // Called when playback stops
    0,                     // Render the PCM data, must return quickly
    avr_matrix_render_freq     // Render the freq data, must return quickly
};



VisPlugin *get_vplugin_info(void)
{
    return &avr_matrix_vtable;
}


static void avr_matrix_init(void) 
{    
    printf("AvrMatrix plugin: Initializing\n");
    avr_matrix_read_config();
    scale = HEIGHT / log(256);
}

static void avr_matrix_cleanup(void)
{    
    printf("AvrMatrix plugin: Cleanup, closing %s\n",avr_matrix_config.device);
    close(fd);
}

static void avr_matrix_about(void)
{
    printf("AvrMatrix plugin: About\n");
}


static void avr_matrix_playback_start(void)
{
    printf("AvrMatrix plugin: Playback starting\n");
    printf("AvrMatrix plugin: Opening %s\n",avr_matrix_config.device);
    struct termios options;
    fd = open( avr_matrix_config.device, O_RDWR | O_NOCTTY | O_NDELAY );
    if (!fd) {
        printf("AvrMatrix plugin: failed opening %s\n",avr_matrix_config.device);
        return;
    }
    fcntl( fd, F_SETFL, 0 );
    tcgetattr(fd, &options);
    cfsetispeed(&options, B115200);
    cfsetospeed(&options, B115200);
    options.c_cflag |= (CLOCAL | CREAD);
    tcsetattr(fd, TCSANOW, &options);
    options.c_cflag &= ~CSIZE;     /*Mask the character size bits*/
    options.c_cflag |= CS8;          /*Select 8 data bits*/
    options.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG);
    options.c_oflag &= ~OPOST;
    options.c_cc[VMIN] = 0;
    options.c_cc[VTIME] = 10;
    printf("AvrMatrix plugin: mode_line %i\n",avr_matrix_config.mode_bar);
    printf("AvrMatrix plugin: mode_freq %i\n",avr_matrix_config.mode_freq);
    printf("AvrMatrix plugin: mode_invert %i\n",avr_matrix_config.mode_invert);
    printf("AvrMatrix plugin: mode_flip %i\n",avr_matrix_config.mode_flip);
    printf("AvrMatrix plugin: mode_old_peak %i\n",avr_matrix_config.mode_hold_peak);
    
    
}

static void avr_matrix_playback_stop(void)
{
    printf("AvrMatrix plugin: Playback stopping closing %s\n",avr_matrix_config.device);
    close(fd);
    fd = -1;
}

static void avr_matrix_send(void)
{
    gint i;
    char val;
    char buffer[12];
    for(i = 0; i < NUM_BANDS; i++){
        if (avr_matrix_config.mode_invert)
            val = ~ ((1 << (int)bar[i][HISTORY_BAR]) - (int)avr_matrix_config.mode_bar) ;
        else
            val = (1 << (int)bar[i][HISTORY_BAR]) - (int)avr_matrix_config.mode_bar  ;
    
        if (avr_matrix_config.mode_hold_peak && peak[i][HISTORY_PEAK] > bar[i][HISTORY_BAR] )
            val |=  (1 << (int)peak[i][HISTORY_PEAK]); 
        
        if (avr_matrix_config.mode_flip){
            char tmp;
            int j;
            tmp = 0;
            for(j = 0; j < 8; j++){
               tmp = (tmp << 1) | (val & 1);
               val >>= 1;
            }
            val = tmp;
        } 
        buffer[i] =val; 
    }
    buffer[NUM_BANDS] = 0x0a;/* set end off packet*/
    if (fd)
        write(fd,buffer,NUM_BANDS+1);
}

static void avr_matrix_render_freq(gint16 data[2][256])
{
    gint i,c,j;
    gint y;
    gfloat sum_bar;
    gfloat sum_peak;
    
    gint * xscale;
    if (avr_matrix_config.mode_freq)
        xscale = xscale1;
    else
        xscale = xscale2;
        
    
    
    for(i = 0; i < NUM_BANDS; i++){
        for(c = xscale[i], y = 0; c < xscale[i + 1]; c++){
            if(data[0][c] > y)
                y = data[0][c];
        }
        y >>= 6;
        if(y != 0){
            y = (gint)(log(y) * scale);
            if(y >= HEIGHT)
                y = HEIGHT;
        }
        sum_bar = sum_peak = 0;
        
        for (j=0;j<HISTORY_BAR;j++)
            sum_bar += bar[i][j];
        sum_bar = sum_bar / HISTORY_BAR;         
        
        for (j=0;j<HISTORY_PEAK;j++)
            sum_peak += peak[i][j];
        sum_peak = sum_peak / HISTORY_PEAK;         
        
        
        if(y > sum_bar){
            bar[i][HISTORY_BAR] = y;
            for (j=0;j<HISTORY_BAR;j++)
                bar[i][j] = y;
            
        } else{
            bar[i][HISTORY_BAR] = sum_bar;         
        }
        
        if(y > sum_peak){
            peak[i][HISTORY_PEAK] = y;
            for (j=0;j<HISTORY_PEAK;j++)
                peak[i][j] = y;
        } else{
            peak[i][HISTORY_PEAK] = sum_peak;         
        }
        for (j=0;j<HISTORY_BAR-1;j++)
            bar[i][j] = bar[i][j+1];
        
        for (j=0;j<HISTORY_PEAK-1;j++)
            peak[i][j] =  peak[i][j+1];
        
        bar[i][HISTORY_BAR-1] = y;
        peak[i][HISTORY_PEAK-1] = y;
        
    }
    avr_matrix_send();
    return;	
}

static void avr_matrix_read_config(void)
{
        ConfigFile *cfg;	
        gchar *filename;

        
        filename = g_strconcat(g_get_home_dir(), "/.xmms/config", NULL);
        cfg = xmms_cfg_open_file(filename);
        avr_matrix_config.device = g_strdup("/dev/ttyS0");
        if (cfg)
        {
                xmms_cfg_read_boolean(cfg, "AVR Matrix", "mode_line", &avr_matrix_config.mode_bar);
                xmms_cfg_read_boolean(cfg, "AVR Matrix", "mode_freq", &avr_matrix_config.mode_freq);
                xmms_cfg_read_boolean(cfg, "AVR Matrix", "mode_invert", &avr_matrix_config.mode_invert);
                xmms_cfg_read_boolean(cfg, "AVR Matrix", "mode_flip", &avr_matrix_config.mode_flip);
                xmms_cfg_read_boolean(cfg, "AVR Matrix", "mode_hold_peak", &avr_matrix_config.mode_hold_peak);
                xmms_cfg_read_string(cfg, "AVR Matrix", "device", &avr_matrix_config.device);

                xmms_cfg_free(cfg);
        }
        g_free(filename);
        
}



static void configure_ok(GtkWidget *w, gpointer data)
{
        ConfigFile *cfg;	
        gchar *filename;
        avr_matrix_config.device = g_strdup(gtk_entry_get_text(GTK_ENTRY(options_dev)));
        avr_matrix_config.mode_bar = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(options_line));
        avr_matrix_config.mode_freq = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(options_freq));
        avr_matrix_config.mode_invert = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(options_invert));
        avr_matrix_config.mode_flip = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(options_flip));
        avr_matrix_config.mode_hold_peak = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(options_peak));

        filename = g_strconcat(g_get_home_dir(), "/.xmms/config", NULL);
        cfg = xmms_cfg_open_file(filename);
        if (!cfg)
            cfg = xmms_cfg_new();
        xmms_cfg_write_boolean(cfg, "AVR Matrix", "mode_line", avr_matrix_config.mode_bar);
        xmms_cfg_write_boolean(cfg, "AVR Matrix", "mode_freq", avr_matrix_config.mode_freq);
        xmms_cfg_write_boolean(cfg, "AVR Matrix", "mode_invert", avr_matrix_config.mode_invert);
        xmms_cfg_write_boolean(cfg, "AVR Matrix", "mode_flip", avr_matrix_config.mode_flip);
        xmms_cfg_write_boolean(cfg, "AVR Matrix", "mode_hold_peak", avr_matrix_config.mode_hold_peak);
        xmms_cfg_write_string(cfg, "AVR Matrix", "device", avr_matrix_config.device);

        xmms_cfg_write_file(cfg, filename);
        xmms_cfg_free(cfg);
        g_free(filename);
        gtk_widget_destroy(configure_win);
}

void avr_matrix_configure (void)
{
        if(configure_win)
                return;

        avr_matrix_read_config();

        configure_win = gtk_window_new(GTK_WINDOW_DIALOG);
        gtk_container_set_border_width(GTK_CONTAINER(configure_win), 10);
        gtk_window_set_title(GTK_WINDOW(configure_win), "AVR Matrix configuration");
        gtk_window_set_policy(GTK_WINDOW(configure_win), FALSE, FALSE, FALSE);
        gtk_window_set_position(GTK_WINDOW(configure_win), GTK_WIN_POS_MOUSE);
        gtk_signal_connect(GTK_OBJECT(configure_win), "destroy", GTK_SIGNAL_FUNC(gtk_widget_destroyed),
                        &configure_win);

        vbox = gtk_vbox_new(FALSE, 5);

        options_frame = gtk_frame_new("Options:");
        gtk_container_set_border_width(GTK_CONTAINER(options_frame), 5);

        options_vbox = gtk_vbox_new(FALSE, 5);
        gtk_container_set_border_width(GTK_CONTAINER(options_vbox), 5);

        
        options_dev = gtk_entry_new();
        gtk_entry_set_text(GTK_ENTRY(options_dev), avr_matrix_config.device);
        gtk_box_pack_start(GTK_BOX(options_vbox), options_dev, FALSE, FALSE, 0);
        gtk_widget_show(options_dev);
        
        
        options_line = gtk_check_button_new_with_label("Bar lines");
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(options_line), avr_matrix_config.mode_bar);
        gtk_box_pack_start(GTK_BOX(options_vbox), options_line, FALSE, FALSE, 0);
        gtk_widget_show(options_line);

        options_freq = gtk_check_button_new_with_label("Wide freq range");
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(options_freq), avr_matrix_config.mode_freq);
        gtk_box_pack_start(GTK_BOX(options_vbox), options_freq, FALSE, FALSE, 0);
        gtk_widget_show(options_freq);
         
        options_invert = gtk_check_button_new_with_label("Invert bar");
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(options_invert), avr_matrix_config.mode_invert);
        gtk_box_pack_start(GTK_BOX(options_vbox), options_invert, FALSE, FALSE, 0);
        gtk_widget_show(options_invert);

        options_flip = gtk_check_button_new_with_label("Flip bar");
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(options_flip), avr_matrix_config.mode_flip);
        gtk_box_pack_start(GTK_BOX(options_vbox), options_flip, FALSE, FALSE, 0);
        gtk_widget_show(options_flip);
        
        options_peak = gtk_check_button_new_with_label("Hold Peak");
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(options_peak), avr_matrix_config.mode_hold_peak);
        gtk_box_pack_start(GTK_BOX(options_vbox), options_peak, FALSE, FALSE, 0);
        gtk_widget_show(options_peak);
                
        gtk_container_add(GTK_CONTAINER(options_frame), options_vbox);
        gtk_widget_show(options_vbox);

        gtk_box_pack_start(GTK_BOX(vbox), options_frame, TRUE, TRUE, 0);
        gtk_widget_show(options_frame);

        bbox = gtk_hbutton_box_new();
        gtk_button_box_set_layout(GTK_BUTTON_BOX(bbox), GTK_BUTTONBOX_END);
        gtk_button_box_set_spacing(GTK_BUTTON_BOX(bbox), 5);
        gtk_box_pack_start(GTK_BOX(vbox), bbox, FALSE, FALSE, 0);

        ok = gtk_button_new_with_label("Ok");
        gtk_signal_connect(GTK_OBJECT(ok), "clicked", GTK_SIGNAL_FUNC(configure_ok), NULL);
        GTK_WIDGET_SET_FLAGS(ok, GTK_CAN_DEFAULT);
        gtk_box_pack_start(GTK_BOX(bbox), ok, TRUE, TRUE, 0);
        gtk_widget_show(ok);
        

        cancel = gtk_button_new_with_label("Cancel");
        gtk_signal_connect_object(GTK_OBJECT(cancel), "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy),
                                GTK_OBJECT(configure_win));
        GTK_WIDGET_SET_FLAGS(cancel, GTK_CAN_DEFAULT);
        gtk_box_pack_start(GTK_BOX(bbox), cancel, TRUE, TRUE, 0);
        gtk_widget_show(cancel);
        gtk_widget_show(bbox);
        
        gtk_container_add(GTK_CONTAINER(configure_win), vbox);
        gtk_widget_show(vbox);
        gtk_widget_show(configure_win);
        gtk_widget_grab_default(ok);
}

